#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h> // for mkdir
#include <inttypes.h>
#include <stdint.h>
#include <algorithm> // for max/min
#include <sys/time.h>
#include <iostream>
#include <errno.h>

#include "minia/Bank.h"
#include "minia/Hash16.h"
#include "minia/Set.h"
#include "minia/Pool.h"
#include "minia/Bloom.h"
#include "minia/Debloom.h"
#include "minia/Utils.h"
#include "minia/SortingCount.h"
#include "minia/Terminator.h"
#include "minia/Kmer.h"
#include "minia/GraphOutput.h"

#include "GraphAnalysis.h"
#include "FindBreak.h"
#include "IterativeExtensions.h"
#include "IndexPairs.h"


#define NVERSION 6417

using namespace std;
int order=0; // in ultimate, don't change it, it should be 0

int64_t genome_size;
int max_insertions_size;
Bloom * bloo1;
BranchingTerminator *terminator;
//extern AssocPairedSet *pairedBranchingKmers;

const string linear_seqs_suffix = ".linearSeqs";
const string insertions_suffix = ".insertions.fa";

const string rright = "right";


int graph_format = 0; // 0 = dot, 1 = kissplice

//Raluca
bool paired;


int identNT(char a, char b)
{
    return   ( (a==b || a-b==32 || a-b ==-32) && a!='N');
}


set< std::pair<int,int> >  find_nodes_containing_R(string R, string linear_seqs_name, int nb_mis_allowed)
{
    bool debug = false;
    //set<int> terminal_nodes;
    set< std::pair<int,int> >  terminal_nodes;
    Bank *Nodes = new Bank((char *)linear_seqs_name.c_str());

    long nodeNb = 0;
    char * nodeseq;
    int nodelen;
    int nbmatch =0;
    
    const char * anchor = R.c_str();
    int anchor_size=  R.size();

    // heuristics: R has to be seen entirely in the node up to nb_mis_allowed errors, in the forward strand

    while (Nodes->get_next_seq(&nodeseq,&nodelen))
    {
        if (nodelen < R.size())
        {
            nodeNb++;
            continue;
        }

        if (debug)
            printf("searching %s (size=%zu nt) in %s (size=%d nt)\n",anchor,R.size(),nodeseq,nodelen);

        for (int j = 0; j < nodelen-R.size()+1; j++)
        {
        
            nbmatch=0;
            
            for (int i = 0; i < anchor_size; i++)
            {
                nbmatch += identNT(nodeseq[j+i], anchor[i]);
            }
            if(nbmatch >= (anchor_size - nb_mis_allowed))
            {
                terminal_nodes.insert( std::make_pair (nodeNb,j) ); // nodeNb,  j pos of beginning of right anchor
                                break;
            }
        }
        nodeNb++;
    }

    if (debug)
    {
        for(set< std::pair<int,int> >::iterator it = terminal_nodes.begin(); it != terminal_nodes.end() ; ++it)
            fprintf(stderr," (node %d pos %d) ",(*it).first,(*it).second);
    }

    Nodes->close();
    delete Nodes;
    return terminal_nodes;
}

/*
 * main gap-filling function:
 *  - construct graph nodes
 *  - construct edges by retrieving (k-1)-overlaps
 *  - find all paths between L and R
 *  - convert them to sequencse
 *  - make sure these sequence are all pair-wise XX% identical
 *
 * note: not thread-safe because GraphOutput as well as find_nodes_containing_R() modify sizeKmer
 */
set<string> gapfill(string L, string R, int estimated_gap_length, int max_gap_length, int verbose, int max_nodes, int nb_mis_allowed,int paired_mode, bool return_all, parcours_t search_mode, bool swf)
{
    set<string> sequences;
    set<string> sequences_upto_R;

    bool verb = verbose>=1 ? 1:0;
    //verb = 1;
    string linear_seqs_name = (prefix + linear_seqs_suffix);
    if(verb) fprintf(stderr," contigs..");
    //IterativeExtensions::traversal_type = IterativeExtensions::SimplePaths;
    //    if(paired_mode)
    //        IterativeExtensions::construct_linear_seqs_paired(L, R, max_gap_length, linear_seqs_name, verbose, max_nodes);
    //    else
    IterativeExtensions::construct_linear_seqs(L, R, max_gap_length, linear_seqs_name, verbose, max_nodes,search_mode,swf);

    terminator->reset();
    // connect the contigs into a graph 

    GraphOutput graph_output = GraphOutput(prefix, graph_format);

    if(verb) fprintf(stderr," linkage..");

    graph_output.load_nodes_extremities(linear_seqs_name);

    graph_output.construct_graph(linear_seqs_name);

    graph_output.close();

    // find R in the nodes 
    // linear_seqs_name contains the nodes in the same order as in the graph

    if(verb)   fprintf(stderr," endpoint..");

  //  set<int> terminal_nodes = find_nodes_containing_R(R, linear_seqs_name, nb_mis_allowed);

    set< std::pair<int,int> > terminal_nodes_with_endpos = find_nodes_containing_R(R, linear_seqs_name, nb_mis_allowed);

    //also convert it to list of node id for traditional use
    set<int> terminal_nodes;
    for (set< std::pair<int,int> >::iterator it = terminal_nodes_with_endpos.begin(); it != terminal_nodes_with_endpos.end(); it++)
    {
        terminal_nodes.insert((*it).first);
    }


    
    if(verb)   printf("nodes found with R  : %zu\n",terminal_nodes.size());
    // analyze the graph to find a satisfying gap sequence between L and R

    GraphAnalysis graph = GraphAnalysis(prefix);
    graph.debug = verbose;

    if(terminal_nodes.size()==0)
    {
        if(verb)  printf("Right anchor not found.. gapfillling failed... \n");
        return sequences;
    }

    if(verb)    fprintf(stderr," analysis..");
    bool success;
    set<unlabeled_path> paths = graph.find_all_paths(terminal_nodes, success);

    
    //commmented this 20 december, should still analyse first paths returned ..
//    if (!success)
//    {
//        if(verb)  printf(" [FAILED] (graph is too complex)\n");
//        return sequences;
//    }

    
    //now this func also cuts the last node just before the beginning of the right anchor
    sequences = graph.paths_to_sequences(paths,terminal_nodes_with_endpos);

    ///////avant analyse si almost identical, couper ce quil y a apres le R
    // sinon si noeud grand, ecart relatif entre 2 inserts est écrasé

    
    ///a faire : faire func pour virer le L et le R ici de sequences_upto_R , en gerant nb mis allowed
    //ptet juste a v
    
    //maintenant R  est vire par paths_to_sequences
    
  //  printf("in gapfill seq nb %lu \n",sequences.size());
 /*   for (set<string>::iterator it = sequences.begin(); it != sequences.end() ; ++it)
    {
        string insertion = *it;
        int R_match = insertion.rfind(R); //todo adapter si n mismatch allowed
        insertion = insertion.substr(0,R_match+R.length()); // inclut le R
        sequences_upto_R.insert(insertion);
      //  printf("insertion %s  strlen %lu \n",insertion.c_str(),strlen(insertion.c_str()));
    }*/

    
//    printf("------------------\n");
//    int iii=0;
//    for (set<string>::iterator it = sequences.begin(); it != sequences.end() ; ++it)
//    {
//
//        printf("seq %i: %s\n",iii,(*it).c_str());
//        iii++;
//    }
//    printf("------------------\n");

    
    
    if (return_all)
        return sequences;
    
    
    //mettre printf si term node found mais pas de chemin trouvé ..
    //et printf si pas de term node found ..
    // check that all sequences are all similar pair-wise
    if (MonumentTraversal::all_consensuses_almost_identical(sequences))
    {
        if(verb)     printf(" [SUCCESS]\n");
        if (sequences.size() > 1)
            sequences.erase(++(sequences.begin()),sequences.end()); // keep only one consensus sequence
    }
    else
        if(verb)   printf(" [MULTIPLE SOLUTIONS]\n");

    return sequences;
}


int main(int argc, char *argv[])
{
    

    
    int nb_default_arguments = 2; // mode and reads/breakpoint file
    int max_disk_space = 0; // let dsk decide

    sizeKmer=27;
    nks = 3;
    strcpy(prefix,"");
    genome_size = 0;
    max_insertions_size = 10000;
    int verbose = 0;
    int max_nodes = 100;
    int nb_mis_allowed = 0;
    int nb_rep_allowed = 5;

    bool correc =false;
    bool findall =false;
    bool paired_mode =false;
    int forward_reverse_mode = 2;
    bool return_all =false;
    bool swf = false;
    bool with_hetero = true;
    parcours_t parcours_mode = LARGEUR; //default is now bfs
    int nb_authorized_kmers_occ_hetero = 1;  // 1 == kmer must be unique
    if(argc < 1+nb_default_arguments)
    {
        fprintf (stderr,"\nMindTheGap usage (<mandatory arguments>, [optional arguments]):\n");
        fprintf (stderr," %s index reads.fa <-g estimated_genome_size> <-p project_name> [-k kmer_size] [-t min_abundance]\n",argv[0]);
        fprintf (stderr," %s find reference.fa <-p project_name>\n",argv[0]);
        fprintf (stderr," %s fill breakpoints.fa <-p project_name> [-i insertion_length]\n",argv[0]);
        fprintf (stderr, "\n");
        fprintf (stderr,"details:\n");
        fprintf (stderr," <-p project_name> any filename prefix for this project\n");
        fprintf (stderr," <-g estimated_genome_size> rough estimation of the length of the genome present in the reads (nt) (index option)\n");
        fprintf (stderr," [-k kmer_size] default: %d (nt); type \"make k=XX\" to increase the maximum supported k value at compile-time. (index option)\n",sizeKmer);
        fprintf (stderr," [-d max_disk_space] is in MB, for index phase,  default: min(available disk space / 2, reads file size) (index option)\n");

        fprintf (stderr," [-t min_abundance] k-mers solidity, default: %d (all kmers seen < %d times are filtered out, index option).\n",nks,nks);
        fprintf (stderr," [-mrep ], maximal repetition size detected for fuzzy site (def 5) (find option)\n");
        fprintf (stderr," [-i insertion_length], default %d; maximum length of insertions (nt) (fill option)\n", max_insertions_size);
        fprintf (stderr," [-n max nodes], default %d; maximum number of nodes in contig graph (nt) (fill option)\n", max_nodes);
        fprintf (stderr," [-m nb mismatches], default %d; maximum number of mismatches in right anchor (nt) (fill option)\n", nb_mis_allowed);
        fprintf (stderr," [-r reverse], default 0 : forward only, 1 : reverse only, 2: both (fill option) \n");
        fprintf (stderr," [-all ], flag to return every possible insert, even if close (fill option)\n");

        fprintf (stderr," [-bfs ], search in breadth-first search (default is bfs) (fill option)\n");
        fprintf (stderr," [-dfs ], search in depth-first search (default is bfs) (fill option)\n");
        
        fprintf (stderr," [-hom ], only search for homozygous breakpoints (find option)\n");

        fprintf (stderr," [-swf ], stop search when found  (will be faster, possibly looses acc. when multiple solutions) (fill option)\n");

        
        fprintf (stderr," [-h nb authorized occurences hetero search], default %d; maximum number of occurences of kmers for hetero anchors (nt) (find option)\n", nb_authorized_kmers_occ_hetero);

        return 0;
    }

//run cmd
    char ** cline = argv;

    
    printf("Program mind the gap version %i run with :\n",NVERSION);
    for(int jj=0; jj< argc; jj++)
     {
        printf("%s ", argv[jj]); /* %s instead of %c and drop [i]. */
    }
    printf("\n\n");
    
    // read pre-saved project file
//    char run_info_filename[1024];
//    sprintf(config_filename,"%s.config",prefix);
//
//        FILE *config = fopen(config_filename,"r");
//               fclose(config);
    
    /////
    
    
    
    for (int n_a = nb_default_arguments + 1; n_a < argc ; n_a++)
    {
        if (strcmp(argv[n_a],"-k")==0)
            sizeKmer = atoi(argv[n_a+1]);

        if (strcmp(argv[n_a],"-t")==0)
            nks = atoi(argv[n_a+1]);
        
        if (strcmp(argv[n_a],"-d")==0)
            max_disk_space = atoi(argv[n_a+1]);

        if (strcmp(argv[n_a],"-p")==0)
            strcpy(prefix,argv[n_a+1]);

        if (strcmp(argv[n_a],"-v")==0)
            verbose = 1;

        if (strcmp(argv[n_a],"-vv")==0)
            verbose = 2;

        if (strcmp(argv[n_a],"-mrep")==0)
        nb_rep_allowed = atoi(argv[n_a+1]);
        
        if (strcmp(argv[n_a],"-c")==0)
            correc = true;
        
        if (strcmp(argv[n_a],"-findall")==0)
            findall = true;
        
        if (strcmp(argv[n_a],"-bfs")==0)
            parcours_mode = LARGEUR;
        
        if (strcmp(argv[n_a],"-dfs")==0)
            parcours_mode = PROFONDEUR;
        
        if (strcmp(argv[n_a],"-swf")==0)
            swf = true;
        
        if (strcmp(argv[n_a],"-paired")==0)
            paired_mode = true;
        
        if (strcmp(argv[n_a],"-hom")==0)
            with_hetero = false;
        
        if (strcmp(argv[n_a],"-all")==0)
            return_all = true;
        
        if (strcmp(argv[n_a],"-g")==0) // genome size
            genome_size = atoll(argv[n_a+1]);

        if (strcmp(argv[n_a],"-i")==0) // max insertion size
            max_insertions_size = atoll(argv[n_a+1]);
        
        if (strcmp(argv[n_a],"-n")==0) // max number of nodes in contig graph
            max_nodes = atoi(argv[n_a+1]);
        
        if (strcmp(argv[n_a],"-h")==0) 
            nb_authorized_kmers_occ_hetero = atoi(argv[n_a+1]);
     
        if (strcmp(argv[n_a],"-r")==0)
            forward_reverse_mode = atoi(argv[n_a+1]);
        
        if (strcmp(argv[n_a],"-m")==0) // max number of mismatches in right anchor
            nb_mis_allowed = atoi(argv[n_a+1]);
        
    }
 
    if(strlen(prefix) == 0)
    {
        printf("missing project name (-p parameter)\n");
        exit(1);
    }
    // read pre-saved project file
    char config_filename[1024];
    sprintf(config_filename,"%s.config",prefix); 
    if (strcmp(argv[1],"fill") == 0 || strcmp(argv[1],"find") == 0 )
    {
        FILE *config = fopen(config_filename,"r");
        if (config == NULL)
        {
            printf("missing config file, please index the reads first\n");
            exit(1);
        }
        fscanf(config, "k %d genome_size %lld", &sizeKmer, &genome_size);
        fscanf(config, "\nsolid_kmers %lli", &nsolids);

        fclose(config);
    }

    // let's make k odd for now, because i havnt thought of how to handle palindromes (dont want to stop on them)
    if (sizeKmer%2==0)
    {
        sizeKmer-=1;
        printf("Need odd kmer size to avoid palindromes. I've set kmer size to %d.\n",sizeKmer);
    }
    if (sizeKmer>(sizeof(kmer_type)*4))
    {
        printf("Max kmer size on this compiled version is %zu\n",sizeof(kmer_type)*4);
        exit(1);
    }

    kmerMask=(((kmer_type)1)<<(sizeKmer*2))-1;
    
    kmerMaskm1=(((kmer_type)1)<<((sizeKmer-1)*2))-1;

    double lg2 = log(2);
    NBITS_PER_KMER = log(16*sizeKmer*(lg2*lg2))/(lg2*lg2); // needed to process argv[5]
   
    if (genome_size == 0)
    {
        printf("missing estimated genome size (-g parameter)\n");
        exit(1);
    }

    // save config file
    if (strcmp(argv[1],"index") == 0  )
    {
        FILE *config = fopen(config_filename,"w");
        fprintf(config, "k %d genome_size %lld\n", sizeKmer, genome_size);
        fclose(config);
    }

    // shortcuts to go directly to assembly using serialized bloom and serialized hash
    int START_FROM_SOLID_KMERS=0; // if = 0, construct the fasta file of solid kmers, if = 1, start directly from that file 
    int LOAD_FALSE_POSITIVE_KMERS=0; // if = 0, construct the fasta file of false positive kmers (debloom), if = 1, load that file into the hashtable
    int NO_FALSE_POSITIVES_AT_ALL=0; // if = 0, normal behavior, if = 1, don't load false positives (will be a probabilistic de bruijn graph)
    for (int n_a = nb_default_arguments; n_a < argc ; n_a++)
    {
        if (strcmp(argv[n_a],"--dont-count")==0)
            START_FROM_SOLID_KMERS = 1;

        if (strcmp(argv[n_a],"--dont-debloom")==0)
            LOAD_FALSE_POSITIVE_KMERS = 1;

        if (strcmp(argv[n_a],"--just-graph")==0)
        {
            START_FROM_SOLID_KMERS = 1;
            LOAD_FALSE_POSITIVE_KMERS = 1;
        }
    }

    STARTWALL(0);

    // index mode
    int max_memory;
    if (strcmp(argv[1],"index") == 0)
    {
        // given genome size, do some estimations
        // int estimated_bloom_size = max( (int)ceilf(log2f(genome_size * NBITS_PER_KMER )), 1);
        uint64_t estimated_bloom_size = (uint64_t) (genome_size * NBITS_PER_KMER);

        uint64_t estimated_nb_FP =  (uint64_t)(genome_size * 4 * powf(0.6,11)); // just indicative
        max_memory =  max ((int64_t)estimated_bloom_size/8LL /1024LL/1024LL, 1LL);

        printf("estimated values: nbits Bloom %lli, nb FP %lld, max memory %i MB\n",estimated_bloom_size,estimated_nb_FP,max_memory);

        Bank *Reads = new Bank(argv[2]);

        // count kmers, write solid kmers to disk
        if (!START_FROM_SOLID_KMERS)
        {
            sorting_count(Reads,prefix,max_memory,max_disk_space,false,verbose);
        }

        delete Reads;

        // debloom, write false positives to disk, insert them into false_positives
        if (! LOAD_FALSE_POSITIVE_KMERS)
        {
            debloom(order, max_memory);
        }

        // load debloom structure to prepare dumps
        bloo1 = bloom_create_bloo1((BloomCpt *)NULL);
        false_positives = load_false_positives();

        // dump bloom and branching kmers once and for all
        bloo1->dump(return_file_name(bloom_file)); 
        BinaryBank *SolidKmers = new BinaryBank(return_file_name(solid_kmers_file),sizeof(kmer_type),0);

        terminator = new BranchingTerminator(SolidKmers,genome_size, bloo1,false_positives);

        BinaryBank *BranchingKmers = new BinaryBank(return_file_name(branching_kmers_file),sizeof(kmer_type),true);

        terminator->dump_branching_kmers(BranchingKmers);


        BranchingKmers->close();


        FILE *config = fopen(config_filename,"a");
        fprintf(config, "solid_kmers %lli\n", SolidKmers->nb_elements());
        fclose(config);

        SolidKmers->close();


        //Branching kmers have been indexed. 
        //Insert code here to compute 'paired kmers'
        
        //Raluca
        //open the reads file, dont't know if it is a fasta/q or a list of file names yet
        //if it's a list then if there's only two files, they may be paired reads
        /*
        if(paired_mode)
        {
        paired = false;
        
        gzFile tempfile = gzopen(argv[2],"r");
        if (tempfile == NULL)
        {
            char buffer[ BUFSIZ ];
            char * errorMessage = (char * ) strerror_r( errno, buffer, BUFSIZ ); // get string message from errno
            printf("error during fopen: %s   %s \n",errorMessage,argv[2]);
            exit(1);
        }
        char deb=(char)gzgetc(tempfile);
        
        char **nfname;
        nfname = (char**) malloc(sizeof(char*)*MAX_NB_FILES);
        for(int jj=0; jj<MAX_NB_FILES; jj++ )
            nfname [jj] =  (char*) malloc(sizeof(char)*TAILLE_NOM);
        
        if(deb=='>' || deb=='@')
        { // file is a fasta/q file, nothing to be done
            gzclose(tempfile);
        }
        else // file contains a list of file names
        {
            char* ret;
            gzungetc(deb,tempfile);
            printf("File %s starts with character \"%c\", hence is interpreted as a list of file names\n",argv[2],deb );
            int ii;
            // get the filenames
            for (ii=0; ii<MAX_NB_FILES ; ii++)
            {
                ret = gzgets(tempfile, nfname[ii], BUFFER_SIZE);
                if (ret != NULL) {
                    // remove \r \n chars
                    char *endline = strchr(nfname[ii], '\n');
                    if (endline)
                        *endline='\0';
                    endline = strchr(nfname[ii], '\r');
                    if (endline)
                        *endline='\0';
                }
                else // no more filenames
                    break;
            }
            
            //if there's only 2 files, then we shall interpret them as paired reads files ... for the moment at least
            if (ii == 2)
            {
                paired = true;
                printf("Reading in parallel the 2 paired reads files %s and %s\n", nfname[0], nfname[1]);
            }
            else
            {
                //nothing to be done, as they're not paired reads files
                paired = false;
            }
            
            gzclose(tempfile);
        }
        
        
        
        ///////////////////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////////////////
        if ( paired )
        {
            pairedBranchingKmers = new AssocPairedSet();
            Bank *Reads1 = new Bank(nfname[1]);
            Bank *Reads2 = new Bank(nfname[0]);
            
            BinaryBank *BranchingKmers = new BinaryBank(return_file_name(branching_kmers_file),sizeof(kmer_type),false);

            pairedBranchingKmers->load_branching(BranchingKmers);
            BranchingKmers->close();

            buildPairedBranchingKmers(pairedBranchingKmers, Reads1, Reads2, bloo1, terminator,false_positives);
            
            //visit all branching kmers with their info associated
           // pairedBranchingKmers->print();
            
            pairedBranchingKmers->dump(return_file_name(assoc_kmer_file));
            printf("\n  -- done dumping branching kmers with paired info\n\n");
        }
            
            
        }// fin paired_mode
        ///////////////////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////////////////
        
         */
        
        STOPWALL(0,"Total");
        printf("\n  -- done indexing reads file (%s), restart in MindTheGap with:\n\n\t%s find genome.fa -p %s\n\n",argv[2],argv[0],prefix);

//        printf("\n  -- done indexing reads file (%s), restart in MindTheGap with:\n\n\t%s fill [breakpoints_file.fa] -p %s\n\n",argv[2],argv[0],prefix);
        

        remove(return_file_name("debloom2"));
        remove(return_file_name(solid_kmers_file));
        remove(return_file_name(binary_read_file));
        
        return 0;

    }
    else
    {
        if (strcmp(argv[1],"fill") != 0 && strcmp(argv[1],"find") != 0)
        {
            printf("unsupported mode: %s\n",argv[1]);
            exit(1);
        }
    }

    // load bloom structure
    bloo1 = bloom_create_bloo1((BloomCpt *)NULL, true);// from dump 

    false_positives = load_false_positives();

    // IterativeExtensions does not load terminator anymore.. so load it here
    Terminator::verbose = false;
    BinaryBank *BranchingKmers = new BinaryBank(return_file_name(branching_kmers_file),sizeof(kmer_type),false);
    terminator = new BranchingTerminator(BranchingKmers,(BinaryBank *)NULL, bloo1,false_positives); //(BinaryBank *) NULL to know which constructoir to call

    if(paired_mode)
    {
        printf("reload pair info \n");
        pairedBranchingKmers = new AssocPairedSet();
        pairedBranchingKmers->load_branching(BranchingKmers);
        pairedBranchingKmers->load(return_file_name(assoc_kmer_file));
        //   pairedBranchingKmers->print();

    }

    BranchingKmers->close();


    if ( strcmp(argv[1],"find") == 0)
    {
        //////-------------------------------------------------------------------------------------------
        fprintf (stderr," Search for discrepancies between genome ref and reads \n");
        
        //call func to find breakpoints
        //and classify breakpoints between insertion, deletion, snp, inversion ...
        Bank *Ref = new Bank(argv[2]);
        
        if(correc)
        {
            printf("Correction mode \n");
            errCorrec(Ref,bloo1,false_positives,NULL);
            return 0;
            STOPWALL(0,"Total");

        }
        else  //find all, with anti  repeat module
        {
            Bloom * bloom_ref_repeats  =NULL;

            ///////--------------test module for het snps : count kmers in ref, put repeated kmers in a bloom ----------------------------------------
            printf("--------count ref--------\n");
            int backup_nks=nks;
            int backup_sizekmer = sizeKmer ;
            sizeKmer = sizeKmer -1; // we count  the k-1 mers
            
            //update kmermask and nbits
            kmerMask=(((kmer_type)1)<<(sizeKmer*2))-1;
            NBITS_PER_KMER = log(16*sizeKmer*(lg2*lg2))/(lg2*lg2);

            
            nks = nb_authorized_kmers_occ_hetero +1 ; // to get the list of non unique kmers on the reference sequence

            uint64_t estimated_bloom_size = (uint64_t) (genome_size * NBITS_PER_KMER);
            max_memory =  max ((int64_t)estimated_bloom_size/8LL /1024LL/1024LL, 1LL);

            printf("doing sorting count with max memory : %i  (read gen size %lli)\n",max_memory,genome_size);
            sorting_count(Ref,prefix,max_memory,max_disk_space,false,verbose);

            BinaryBank *repeated_kmers = new BinaryBank(return_file_name(solid_kmers_file),sizeof(kmer_type),0);
            uint64_t nb_repeats =repeated_kmers->nb_elements();

            printf("%lli repeated kmers in the ref sequence -->  bloom %lli MB\n",nb_repeats ,(uint64_t)(nb_repeats*NBITS_PER_KMER /1024LL/1024LL));
            if(nb_repeats==0)nb_repeats=1;

            bloom_ref_repeats  = bloom_create_bloo1((BloomCpt *)NULL);

            //  Bloom * bloom_ref_repeats  = new Bloom((uint64_t)nb_repeats*NBITS_PER_KMER);
            //     bloo1->set_number_of_hash_func((int)floorf(0.7*NBITS_PER_KMER));
            nks=backup_nks;
            sizeKmer = backup_sizekmer ;
            
            kmerMask=(((kmer_type)1)<<(sizeKmer*2))-1;
            NBITS_PER_KMER = log(16*sizeKmer*(lg2*lg2))/(lg2*lg2);
            
            remove(return_file_name(solid_kmers_file));
            printf("end count ref\n");


            Ref->rewind_all();

            ///////----------------------------------------end ----------------------------------------

            if(findall)
            {
                printf("Find eveything  \n");
                findAllBreakpoints(Ref,bloo1,false_positives,bloom_ref_repeats);

            }
            else
            {
                printf("Find insert sites  \n");

                findbreakpoints(Ref,bloo1,false_positives,bloom_ref_repeats,with_hetero,nb_rep_allowed);
            }
            
            printf("\n  -- done searching breakpoints, restart in MindTheGap with:\n\n\t%s fill %s.breakpoints -p %s\n\n",argv[0],prefix,prefix);
            STOPWALL(0,"Total");

            return 0;

        }


    }

    //////-------------------------------------------------------------------------------------------
    fprintf (stderr,"______________________________________________________ \n");
    fprintf (stderr,"______________________ [ultimate] NNNN [gapfiller] ___ \n");
    fprintf (stderr,"______________________________________________________ \n\n");
    //////-------------------------------------------------------------------------------------------
    
    // load breakpoints

    
    printf("max nodes %i \n",max_nodes);
    Bank *Breakpoints = new Bank(argv[2]);

    string results_filename(prefix);
    results_filename += insertions_suffix; 
    FILE * file_insertions = fopen(results_filename.c_str(),"w+");

    char *buffer ;
    char temp[1000];
    char * tempseq = (char *) malloc(max_insertions_size *sizeof(char));
    int readlen, readlen2;

    char *headerbuffer;
    int hlen;
    
    bool more_to_fill;
    Progress progress;
    
    int64_t nbgaps = Breakpoints->estimate_nb_reads()/2;
    char message[1000];
   // sprintf(message,"Ultimate gapfiller in action ...  %i  gaps to fill",nbgaps);
    sprintf(message,"Ultimate gapfiller in action ...  ");

    if(verbose==0) progress.init(nbgaps,message);
    progress.timer_mode=1;
    more_to_fill = Breakpoints->get_next_seq(&buffer,&headerbuffer,&readlen2,&hlen);
    
    
    int hetmode = 0;
 //   char * res = strstr(headerbuffer,"HET");
//    if(res!=NULL)
//    {
//        hetmode = 1;
//    }

    
   // printf("first buffer 2 %s  more_to_fill %i \n",buffer,more_to_fill);
   // string breakpoint_id(Breakpoints->Seq[0]);
    string breakpoint_id(headerbuffer);
    if (breakpoint_id.length() > 0)
    {
        // remove '>' and '\n'
       breakpoint_id = breakpoint_id.substr(breakpoint_id.find_first_not_of(" >"));
       breakpoint_id = breakpoint_id.substr(0,breakpoint_id.find_first_of("\n\r"));
    }

    string L(buffer);

    //auto adjust starter size
    if(L.size() > sizeKmer)
    {
        L = L.substr(L.size()- sizeKmer,sizeKmer);
    }
    int dd=0;
    while (more_to_fill)
    {
        //printf("gapfill %i %s \n",dd++,L.c_str() );
        // retrieve left and right flanking regions
        
        more_to_fill = Breakpoints->get_next_seq(&buffer,&readlen);
        string R(buffer);
        if(R.size() > sizeKmer)
        {
            R = R.substr(0,sizeKmer);
        }
      //  printf("\nGapfilling breakpoint: \"%s\" \n",breakpoint_id.c_str());
      //  printf("L %s R %s \n",L.c_str(),R.c_str());

        set<string> sequences;
        int nbContig;
        //  printf("filling %s\n",breakpoint_id.c_str());
        if(forward_reverse_mode != 1 ) // 1 is rev only
        {
            sequences = gapfill(L,R,0,max_insertions_size,verbose, max_nodes, nb_mis_allowed,paired_mode, return_all, parcours_mode, swf); // main gap-filling code
            
            // save sequences to results file
          //  printf("nb seq %lu \n",sequences.size());

            //
            nbContig = 0;
            for (set<string>::iterator it = sequences.begin(); it != sequences.end() ; ++it)
            {
                string insertion = *it;
//                int R_match = insertion.rfind(R); //todo modif si nb mis
//                insertion = insertion.substr(0,R_match+R.length()); // pour virer ce quil y a apres le R
                int llen = insertion.length() ;// - (int) R.length() - (int) L.length() - 2*hetmode;
              //  printf("%i llen  %s \n",llen,insertion.c_str());
                
                //discards insert too long (can happen when last node is very large)
                if(llen > max_insertions_size)
                    continue;
                
                if(llen > 0)
                {
                    fprintf(file_insertions,"> insertion %d ( len= %d ) for breakpoint \"%s\"\n",nbContig++,llen,breakpoint_id.c_str());
                    fprintf(file_insertions,"%.*s\n",(int)llen,insertion.c_str() ); //+ L.length()+hetmode
                    // insertion.c_str() + L.length()  pour virer le L  et llen pour virer le R
                }
            }
            
            
        }
        
      //  printf("Filter seq tai <0  seq %lu \n",sequences.size());
        //ici faire filtre pour virer les seq de taille <0
        int iii = 0;
        for (set<string>::iterator it = sequences.begin(); it != sequences.end() ; ++it)
        {
            string insertion = *it;

            if( insertion.length() <= 0)
            {
              //  printf ("seq %i  len %i \n",iii++,insertion.length() );

                             //   sequences.erase(it); // pourquoi ca fait planter sur mac

            }
                
        }
        
        /////if filling both ways, do here the other way, only if insert not found ?
        
        if(forward_reverse_mode != 0  && sequences.size() ==0 ) // 0 is forward only
        {

            strcpy(temp,L.c_str());
            revcomp_sequence(temp,L.length());
            L = string(temp);
            
            strcpy(temp,R.c_str());
            revcomp_sequence(temp,R.length());
            R = string(temp);
            
            R.swap(L);
            
          //  printf("L %s R %s ",L.c_str(),R.c_str());

            sequences = gapfill(L,R,0,max_insertions_size,verbose, max_nodes, nb_mis_allowed,paired_mode, return_all, parcours_mode, swf); // main gap-filling code
            
            // save sequences to results file
            //
            nbContig = 0;
            for (set<string>::iterator it = sequences.begin(); it != sequences.end() ; ++it)
            {
                string insertion = *it;
//                int R_match = insertion.rfind(R); //todo modif si nb mis
//                insertion = insertion.substr(0,R_match+R.length());
                int llen = insertion.length()  ; //- (int) R.length() - (int) L.length() - 2*hetmode ;

                //discards insert too long
                if(llen > max_insertions_size)
                    continue;
                
                if(llen > 0)
                {
                    
                  //  strncpy(tempseq,insertion.c_str()+ L.length(),insertion.length()-R.length() -L.length() );
                    strncpy(tempseq,insertion.c_str(),insertion.length());

                    revcomp_sequence(tempseq,insertion.length());
                    
                    fprintf(file_insertions,"> insertion %d ( len= %d ) for breakpoint \"%s\"\n",nbContig++,llen,breakpoint_id.c_str());
                 //   fprintf(file_insertions,"%.*s\n",(int)(insertion.length()-R.length() -L.length()),insertion.c_str()+ L.length());
                    fprintf(file_insertions,"%.*s\n",(int)llen,tempseq);

                }
            }
            
        }
        
        
        
        ///////////////////////////////////
        
        
        //prepare for next breakpoint
        more_to_fill = Breakpoints->get_next_seq(&buffer,&headerbuffer,&readlen2,&hlen);
        
        hetmode = 0;
      //  char * res = strstr(headerbuffer,"HET");
//        if(res!=NULL)
//        {
//            hetmode = 1;
//        }
        
        L = buffer;
        if(L.size() > sizeKmer)
        {
            L = L.substr(L.size()- sizeKmer,sizeKmer);
        }
        
        //breakpoint_id = Breakpoints->Seq[0];
        breakpoint_id = headerbuffer;
        
        if (breakpoint_id.length() > 0)    {
            // remove '>' and '\n'
            breakpoint_id = breakpoint_id.substr(breakpoint_id.find_first_not_of(" >"));
            breakpoint_id = breakpoint_id.substr(0,breakpoint_id.find_first_of("\n\r"));
        }
        
        
        if(verbose==0) progress.inc(1);
    }
   if(verbose==0) progress.finish();
    fclose(file_insertions);
    fprintf(stderr,"\n");
    
    //////-------------------------------------------------------------------------------------------
    
    STOPWALL(0,"Total");
    printf("\n  -- done filling gaps in breakpoint file (%s), results are in:\n\n\t%s.insertions.fa\n\n",argv[2],prefix);

    return 0;
}


